// *******************************************************************
//   (C) Copyright 2013 Leiden Institute of Advanced Computer Science
//   Universiteit Leiden
//   All Rights Reserved
// *******************************************************************
// Retrograde
// *******************************************************************
// FILE INFORMATION:
//   File:     main.cc
//   Author:   Jan N. van Rijn
//   Revision: 0.01a
//   Date:     2013/02/25
// *******************************************************************
// DESCRIPTION: Retrograde. Builds an endgame database.
// *******************************************************************

#include "../core/move.h"
#include "../core/position.h"

#include "../interface/cstdio_interface.h"
#include "../interface/options.h"

#include "constants.h"
#include "output_writer.h"
#include "history.h"

#include <cstdlib>
#include <iostream>
#include <string>

#include <vector>
#include <map>
#include <algorithm>

using namespace std;

// information which configurations we have processed
bool configurations[CONFIGURATIONS][CONFIGURATIONS];
// history of a configuration
History history[HISTORY_SIZE];
// output writer
Output_writer output;

void reset_configurations() {
  for(int i = 0; i < CONFIGURATIONS; ++i) {
	for(int j = 0; j < CONFIGURATIONS; ++j) {
	  configurations[i][j] = false;
	}
  }
}

void reset_history() {
  for(int i = 0; i < HISTORY_SIZE; ++i) {
    history[i] = History();
  }
}

int compare (const void * a, const void * b) {
  return ( *(int*)a - *(int*)b );
}

int configuration_index(uint32_t strength[USED_PIECES],bool is_black) {
  int index = 0;
  int pieces_in = 0;
  qsort(strength,USED_PIECES,sizeof(uint32_t),compare);
  for(int i = 0; i < USED_PIECES; ++i) {
    if(strength[i] % 2 == is_black) {
	  index += (strength[i] / 2) << (pieces_in * 3);
	  pieces_in += 1;
	}
  }
  return index;
}

bool valid_encounter(uint32_t const strength[USED_PIECES],int index,bool recording) {
  vector<int> positions;
  bool black_piece = false; // checks whether both sides have at least one piece.
  bool white_piece = false;
  
  for(int i = 0; i < USED_PIECES; ++i) {
    uint32_t location = index % BOARD_SIZE;
    index /= BOARD_SIZE;
    if(location != CAPTURED) {
      if(strength[i] > 1 && is_not_water(location) == false) 
        return false;
      if(location == LOCATION_WHITE_DEN || location == LOCATION_BLACK_DEN)
        return false;
      
      if(binary_search(positions.begin(), positions.end(), static_cast<int>(location))) {
        return false;
      } else {
        positions.push_back(location);
        sort(positions.begin(), positions.end());
      }
      
      if(strength[i]%2 == 0)
        white_piece = true;
      else
        black_piece = true;
      
    } else if(recording) {
	  return false;
	}
  }
  return (black_piece && white_piece);
}

int to_history_index(Position & board,
                     uint32_t const strength[USED_PIECES]) {
  int total = board.is_white_turn() ? 0 : HISTORY_HALF;
  for(int i = 0; i < USED_PIECES; ++i) {
    total += board.piece(strength[i]) * (1 << (6 * i));
  }
  return total;
}

void set_pieces(Position & board,
                uint32_t const strength[USED_PIECES], 
                int index) {
  for(int i = 0; i < USED_PIECES; ++i) {
    int location = index % BOARD_SIZE;
    index /= BOARD_SIZE;
    board.set_piece(strength[i],location);
  }
}

uint32_t retrograde(uint32_t strength[USED_PIECES]) {
  uint32_t winning = 0;
  uint32_t losing = 0;
  
  Position board;
  
  for(int i = 0; i < HISTORY_SIZE; ++i) {
	board.clear();
    if( i >= HISTORY_HALF ) board.do_null_move(); // black's turn
    bool const black_to_move = (i >= HISTORY_HALF);
    bool const we = black_to_move;
    bool const them = !black_to_move;
    
    if(valid_encounter(strength,i,false) && !history[i].calculated) {
      bool can_win = false;
      bool can_draw = false;
      
      int best_winning_sequence = HIGH_CONSTANT;
      int best_counter_play = 0;
	  
	  Move move_winning_sequence = Move(NONE, CAPTURED, CAPTURED, NONE);
	  Move move_counter_play = Move(NONE, CAPTURED, CAPTURED, NONE);
      
      Move moves[MAX_MOVES];
      set_pieces(board,strength,i);
      int const count = board.generate_moves(moves);
      
      for(int k = 0; k < count; ++k) { 
        board.do_move(moves[k]);
        int index = to_history_index(board,strength);
        
        if(board.is_won()) {
          // win in 1-ply
          if(best_winning_sequence > 1) {
            // also a new optimum
            best_winning_sequence = 1;
            move_winning_sequence = moves[k];
            can_win = true;
          }
        } else if(history[index].calculated && history[index].winner == we) {
          // win in > 1 ply
          if(history[index].nr_of_moves + 1 < best_winning_sequence) {
            // new optimum
            best_winning_sequence = history[index].nr_of_moves + 1;
            move_winning_sequence = moves[k];
            can_win = true;
          }
        } else if(history[index].calculated == false) {
          can_draw = true;
        } else {
          if(history[index].nr_of_moves + 1 > best_counter_play) {
            best_counter_play = history[index].nr_of_moves + 1;
            move_counter_play = moves[k];
          }
        }
        board.undo_move(moves[k]);
      }
      
      if(can_win) {
        ++winning;
        history[i] = History(we,best_winning_sequence,move_winning_sequence);
      } else if(can_draw == false) {
        ++losing;
        history[i] = History(them,best_counter_play,move_counter_play);
      }
    }
  }
  return winning + losing;
}

void print_data(uint32_t strength[USED_PIECES],bool inversed) {
  int start = 0;
  int end = HISTORY_HALF;
  if(inversed) {
	start = HISTORY_HALF;
	end = HISTORY_SIZE;
  }
  
  for(int i = start; i < end; ++i) {
    if(valid_encounter(strength,i,true)){    
      output.write_line(history[i],strength,i,inversed);
    }
  }
  output.flush_statistics(strength,USED_PIECES);
  
}

void calculate_all(uint32_t strength[USED_PIECES],uint32_t const threshold) {
  int black_pieces = configuration_index(strength,true);
  int white_pieces = configuration_index(strength,false);
  int lowest  = min(black_pieces,white_pieces);
  int highest = max(black_pieces,white_pieces);

  if(!configurations[highest][lowest]) {
	reset_history();
	for( uint32_t i = 0, found = HIGH_CONSTANT; found != 0 && i < threshold; ++i) {
	  found = retrograde(strength);
	}
	
	print_data(strength,false);
	if(black_pieces != white_pieces) {
	  print_data(strength,true);
	}
	cerr << "done with an iteration: " << output.strength_to_string(strength,USED_PIECES) << endl;
	configurations[highest][lowest] = true;
  } else {
	cerr << "skipping an iteration: " << output.strength_to_string(strength,USED_PIECES) << endl;
  }
}

int main(int argc, char* argv[])
{
  char const* threshold_string = get_option(argc, argv, 't');
  char const* white_pieces_string = get_option(argc, argv, 'w');
  char const* black_pieces_string = get_option(argc, argv, 'b');
  
  uint32_t threshold = threshold_string == 0 ? HIGH_CONSTANT : atoi(threshold_string);
  
  reset_configurations();
  clog << "memory_size: " << HISTORY_SIZE << endl;
  clog << "used pieces: " << USED_PIECES << endl;
  
  if (white_pieces_string != 0 && black_pieces_string != 0) {
    string white = string(white_pieces_string);
    string black = string(black_pieces_string);
    transform(white.begin(),white.end(),white.begin(), ::toupper);
    transform(black.begin(),black.end(),black.begin(), ::tolower);
    
    string const pieces = white + black;
    uint32_t strength[USED_PIECES];
    if(pieces.length() != (uint32_t) USED_PIECES) {
      cerr << "Program compiled for use with " << USED_PIECES << "animals. "
           << "Pass this number of characters, or recompile with different " 
           << "value for constant 'USED_PIECES'. " << endl;
      return 1;
    }
    for(int i = 0; i < USED_PIECES; ++i) {
      strength[i] = get_piece(pieces[i]);
    }
    qsort(strength,USED_PIECES,sizeof(uint32_t),compare);
    
    calculate_all(strength,threshold);
    
  } else if (white_pieces_string != 0 || black_pieces_string != 0) {
    cerr << "Program can only be executed if {neither,both} " 
         << "-w and -b arguments are given." << endl;
    return 1;
    
  } else {
    if(USED_PIECES == 2) {
      for(int i = 0; i < TOTAL_PIECES; i+=2) {
        for(int j = 1; j < TOTAL_PIECES; j+=2) {
          uint32_t strength[] = {i,j};
          calculate_all(strength,threshold);
        }
      }
    } else if(USED_PIECES == 3) {
      for(int i = 0; i < TOTAL_PIECES; i+=2) {
        for(int j = 1; j < TOTAL_PIECES; j+=2) {
          for(int k = i+2; k < TOTAL_PIECES; k+=2) {
            uint32_t strength[] = {i,j,k};
            qsort(strength,USED_PIECES,sizeof(uint32_t),compare);
            calculate_all(strength,threshold);
          }
		  for(int k = j+2; k < TOTAL_PIECES; k+=2) {
            uint32_t strength[] = {i,j,k};
            qsort(strength,USED_PIECES,sizeof(uint32_t),compare);
            calculate_all(strength,threshold);
          }
        }
      }
    } else if(USED_PIECES == 4) {
      for(int i = 0; i < TOTAL_PIECES; i+=2) {
        for(int j = 1; j < TOTAL_PIECES; j+=2) {
		  // 2 vs 2
          for(int k = i+2; k < TOTAL_PIECES; k+=2) {
            for(int l = j+2; l < TOTAL_PIECES; l+=2) {
              uint32_t strength[] = {i,j,k,l};
              qsort(strength,USED_PIECES,sizeof(uint32_t),compare);
              calculate_all(strength,threshold);
            }
          }
		  // 3 vs 1
		  for(int k = i+2; k < TOTAL_PIECES; k+=2) {
            for(int l = k+2; l < TOTAL_PIECES; l+=2) {
			  uint32_t strength[] = {i,j,k,l};
              qsort(strength,USED_PIECES,sizeof(uint32_t),compare);
              calculate_all(strength,threshold);
			}
		  }
        }
      }
    } else {
      cerr << "Program does not support other than {2,3,4} pieces. Yet." << endl;
      return 1;
    }
  }
  return 0;
} // main

